import numpy as np
import cv2
import os
import torchvision.transforms as transforms
import torch
import random


def standup_roulette(persons):
    # Select a random person from the list
    random_person = random.choice(persons)

    # Select a random direction
    direction = random.choice(['clockwise', 'counterclockwise'])
    return random_person, direction

def resize_images(images, resize_size):
    resized_images = []

    for img in images:
        current_height, current_width, _ = img.shape

        # Calculate the new dimensions to be 5 times divisible by 2
        new_width = int(np.ceil(current_width / 2 ** 5) * 2 ** 5)
        new_height = int(np.ceil(current_height / 2 ** 5) * 2 ** 5)

        # Check if either dimension is smaller than 128
        if new_width < resize_size or new_height < resize_size:
            # Calculate the scaling factor to ensure both dimensions are at least 80
            scaling_factor = max(resize_size / new_width, resize_size / new_height)
            # Upscale both dimensions with the same factor
            new_width = int(new_width * scaling_factor)
            new_height = int(new_height * scaling_factor)

        scaling_factor_width = current_width / new_width
        scaling_factor_height = current_height / new_height

        # Resize the image using OpenCV or any other library of your choice
        resized_img = cv2.resize(img, (new_width, new_height))
        resized_images.append([resized_img, current_height, current_width, scaling_factor_height, scaling_factor_width])

    return resized_images


def resize_images_tensor(images, resize_size):
    resized_images = []
    resized_images_total = []
    for index, img in enumerate(images[0][0]):

        # Convert PyTorch tensor to NumPy array
        img_np = img.cpu().numpy()
        current_height, current_width, _ = img_np.shape

        # Calculate the new dimensions to be 5 times divisible by 2
        new_width = 96  # int(np.ceil(current_width / 2 ** 5) * 2 ** 5)
        new_height = 128  # int(np.ceil(current_height / 2 ** 5) * 2 ** 5)

        # Check if either dimension is smaller than 128
        if new_width < resize_size or new_height < resize_size:
            # Calculate the scaling factor to ensure both dimensions are at least 80
            scaling_factor = max(resize_size / new_width, resize_size / new_height)
            # Upscale both dimensions with the same factor
            new_width = int(new_width * scaling_factor)
            new_height = int(new_height * scaling_factor)

        # Resize the image using OpenCV or any other library of your choice
        resized_img = cv2.resize(img_np, (new_width, new_height))

        # Convert the resized image back to a PyTorch tensor
        resized_img_tensor = torch.from_numpy(resized_img).cuda()

        # Append information along with the resized image to the list
        resized_images.append(resized_img_tensor)

    resized_images_total.append(resized_images)
    resized_images_total.append(images[1])  # append box coor
    resized_images_total.append(images[2])  # append score
    return resized_images_total


def show_face(frame, resized_faces, start_person, direction):

    # Plot bounding boxes
    for index, faces in enumerate(resized_faces[1][0]):
        if faces is not None:
            # Find the person with the highest score per box
            max_key = None
            max_value = float('-inf')
            try:
                resized_faces[3 + index]
            except:
                print('h')
            for d in resized_faces[3+index]:
                # Find the key with the maximum value in the current dictionary
                current_max_key = max(d, key=d.get)
                current_max_value = d[current_max_key]

                # Update the overall maximum key and value if necessary
                if current_max_value > max_value:
                    max_key = current_max_key
                    max_value = current_max_value

            color = (0, 0, 255)
            text = f"'{max_key}'"
            if start_person is max_key:
                color = (0, 255, 0)
                text = f"'{max_key}' + {str(direction)}"

            face_x1, face_y1, face_x2, face_y2 = faces
            cv2.rectangle(frame, (round(face_x1), round(face_y1)), (round(face_x2), round(face_y2)), color, 2)

            cv2.putText(frame, text, (round(face_x1), round(face_y1) - 10), cv2.FONT_HERSHEY_SIMPLEX, 0.5, color, 2)

        else:
            continue

    # out.write(frame)
    cv2.imshow('Webcam output', frame)


def imread_templates(folder_path):
    image_dict = {}
    # Check if the folder exists
    if os.path.exists(folder_path) and os.path.isdir(folder_path):
        # List all files in the folder
        files = os.listdir(folder_path)

        # Filter only files with specific extensions (e.g., '.png', '.jpg')
        image_files = [f for f in files if f.lower().endswith(('.png', '.jpg', '.jpeg'))]

        # Read each image and append it to the list
        for image_file in image_files:
            image_path = os.path.join(folder_path, image_file)
            image = cv2.imread(image_path)
            image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
            image = transforms.ToTensor()(image).unsqueeze(0).cuda()
            image_name = image_file.split('.')[0]
            if image is not None:
                image_dict[image_name] = image
            else:
                print(f"Failed to read image: {image_path}")
    else:
        print("Folder not found.")
    return image_dict




import cv2
from standup_face_recognition.MTCNN_detector import MTCNN_detector
from face_recognition import Siamese
from utils import imread_templates, resize_images_tensor, show_face, standup_roulette


if __name__ == '__main__':
    # template_dict = imread_templates('/home/timo/face_recognition/team_images')
    names = ['Nitin', 'Timo', 'Hiep', 'Robert', 'Martin', 'Kai', 'Bharat', 'Markus', 'Karl', 'Matthias']
    order_person = ['Timo', 'Bharat', 'Martin', 'Matthias', 'Kai', 'Robert', 'Nitin', 'Hiep', 'Karl']
    person, direction = standup_roulette(names)
    person = 'Karl'
    # face recognition
    face_recognition = Siamese()
    # Open a connection to the webcam (0 represents the default camera)
    cap = cv2.VideoCapture(0)
    # fourcc = cv2.VideoWriter_fourcc(*'XVID')
    # out = cv2.VideoWriter('/home/timo/face_recognition/output.mp4', fourcc, 20.0, (640, 480))

    # Check if the webcam opened successfully
    if not cap.isOpened():
        print("Error: Could not open webcam.")
        exit()

    counter = 0
    smooth_filter = 5
    # moving_avg_faces = []
    moving_avg_faces = 0.0
    while True:
        # Capture frame-by-frame
        ret, frame = cap.read()
        counter += 1
        if counter % smooth_filter == 0:
            counter = 1
            moving_avg_faces = 0.0

        # Check if the frame was successfully captured
        if not ret:
            print("Error: Could not read frame.")
            break

        # MTCNN face detector
        MTCNN_face_detector = MTCNN_detector()
        detected_faces_mtcnn = MTCNN_face_detector.get_bbox_detection(frame)

        if detected_faces_mtcnn[0][0] is not None:
            # resize for face_recognition model
            resized_faces = resize_images_tensor(detected_faces_mtcnn, 128)

            moving_avg_faces = [0.0] * len(resized_faces[0])
            for ind, tensor_faces in enumerate((resized_faces[0])):
                moving_avg_faces[ind] = (moving_avg_faces[ind] + tensor_faces) / counter

            if counter % (smooth_filter-1) == 0:
                avg_faces = [moving_avg_faces]
                avg_faces.extend(resized_faces[1:])
                # MTCNN_face_detector.visualize_face_detection(frame, resized_faces)qn
                face_det_reg = face_recognition.face_recognition(avg_faces, names)

                show_face(frame, face_det_reg, person, direction)
        else:
            cv2.imshow('Webcam output', frame)

        # cv2.imwrite('/home/timo/face_recognition/webcam_images/frame.png', frame)
        # Break the loop if 'q' key is pressed
        if cv2.waitKey(1) & 0xFF == ord('n'):
            person_ind = order_person.index(person)
            if direction is 'clockwise':
                person = order_person[(person_ind + 1) % len(order_person)]
            else:
                person = order_person[(person_ind - 1) % len(order_person)]

        if cv2.waitKey(1) & 0xFF == ord('q'):
            break

    # Release the webcam and close all OpenCV windows
    cap.release()
    # out.release()
    cv2.destroyAllWindows()



from .models.inception_resnet_v1 import InceptionResnetV1
from .models.mtcnn import MTCNN, PNet, RNet, ONet, prewhiten, fixed_image_standardization
from .models.utils.detect_face import extract_face
from .models.utils import training

import warnings
warnings.filterwarnings(
    action="ignore", 
    message="This overload of nonzero is deprecated:\n\tnonzero()", 
    category=UserWarning
)


import cv2
import numpy as np
import torch
from facenet_pytorch import MTCNN


class MTCNN_detector:
    def __init__(self):
        self.detector = MTCNN(thresholds=[0.7, 0.7, 0.7], keep_all=True, device='cuda').eval()
        self.results_list = []

    @staticmethod
    def _pre_processing_mtcnn(image):
        image = np.expand_dims(image, axis=0)
        image = np.float32(image)
        image = torch.from_numpy(image)
        image = image.to(torch.device("cuda"))
        return image

    def get_bbox_detection(self, image):
        input_img = self._pre_processing_mtcnn(image)
        faces, _, box, _, probs = self.detector(input_img)
        result_tuple = [faces, box, probs]
        return result_tuple

    @staticmethod
    def visualize_face_detection(frame, detected_faces):

        # Plot bounding boxes
        for index, faces in enumerate(detected_faces[1][0]):
            if faces is not None:
                # for face in faces[2][0]: # uncomment if all face boxes should be visualized
                face_x1, face_y1, face_x2, face_y2 = faces
                cv2.rectangle(frame, (round(face_x1), round(face_y1)), (round(face_x2), round(face_y2)), (0, 255, 0), 2)

            else:
                continue

        cv2.imwrite('/home/timo/face_recognition/webcam_images/' + 'detected_face_' + str(index) + '.png', frame)



import torch
import torch.nn as nn
from facenet_pytorch import InceptionResnetV1


class Siamese:
    def __init__(self):
        Siamese_network = InceptionResnetV1(pretrained='vggface2')
        checkpoint = torch.load("20180402-114759-vggface2.pt")
        Siamese_network.load_state_dict(checkpoint)
        Siamese_network = Siamese_network.cuda()
        Siamese_network.eval()
        self.Siamese_network = Siamese_network
        self.cos = nn.CosineSimilarity(dim=1, eps=1e-6)
        self.team_embedding = torch.load("team_embedding.pth")

    def face_recognition(self, detected_face, names):

        for det_face in detected_face[0]:
            similarity_list = []
            # embedding_list = []
            embedding1 = self.Siamese_network(det_face.permute(2, 0, 1).unsqueeze(0))
            for index, name in enumerate(names):
                # embedding2 = self.Siamese_network(temp[1])
                # embedding_list.append(embedding2)
                cosine_similarity = self.cos(embedding1, self.team_embedding[index]).item()
                similarity_list.append({name: cosine_similarity})
            # torch.save(embedding_list, '/home/timo/face_recognition/team_images/team_embedding.pth')
            detected_face.append(similarity_list)
        return detected_face



import os
from setuptools import setup, find_packages
from setuptools.command.build_py import build_py


#class build_py_with_pth_file(build_py):
     #"""Include the .pth file for this project, in the generated wheel."""

     #def run(self):
         #super().run()

         #destination_in_wheel = "20180402-114759-vggface2.pt"
         #location_in_source_tree = "standup_face_recognition/20180402-114759-vggface2.pt"
 
         #outfile = os.path.join(self.build_lib, destination_in_wheel)
         #self.copy_file(location_in_source_tree, outfile, preserve_mode=0)


         #destination_in_wheel = "team_embedding.pth"
         #location_in_source_tree = "standup_face_recognition/team_embedding.pth"
 
         #outfile = os.path.join(self.build_lib, destination_in_wheel)
         #self.copy_file(location_in_source_tree, outfile, preserve_mode=0)
         
         

print('find packages: ', find_packages())
setup(
    name='standup-face-recognition',
    version='1.3',
    description='Standup helper: Detects and recognizes the person in the team.',
    author='Timo',
    packages=find_packages(),
    # include_package_data=True,
    # cmdclass={"build_py": build_py_with_pth_file},
    install_requires=[
    	'opencv-python',
    	'numpy',
    	'torch==2.0.0',
    	'torchvision==0.15.1',
    ],
    entry_points={
        'console_scripts': [
            'standup_face_recognition=standup_face_recognition.main:main',
        ],
    },
)
